/* NickServ functions. */

#include "services.h"

static NickInfo *nicklists[256];	/* One for each initial character */

const char s_NickServ[] = "NickServ";

#include "ns-help.c"


/* Definitions for the timeout list: */
typedef struct timeout_ Timeout;
struct timeout_ {
    Timeout *next, *prev;
    NickInfo *ni;
    time_t timeout;
    int type;
};
#define TO_COLLIDE	0	/* Collide the user with this nick */
#define TO_RELEASE	1	/* Release a collided nick */

Timeout *timeouts = NULL;


static int is_on_access(User *u, NickInfo *ni);
static void alpha_insert_nick(NickInfo *ni);
static NickInfo *makenick(const char *nick, const char *pass);
static int delnick(NickInfo *ni);

static void collide(NickInfo *ni);
static void release(NickInfo *ni);

static void add_timeout(NickInfo *ni, int type, time_t delay);
static void del_timeout(NickInfo *ni, int type);

static void do_help(const char *source);
static void do_register(const char *source);
static void do_identify(const char *source);
static void do_drop(const char *source);
static void do_set(const char *source);
static void do_set_password(NickInfo *ni, char *param);
static void do_set_kill(NickInfo *ni, char *param);
static void do_access(const char *source);
static void do_info(const char *source);
static void do_list(const char *source);
static void do_recover(const char *source);
static void do_release(const char *source);
static void do_ghost(const char *source);
static void do_getpass(const char *source);

/*************************************************************************/

/* Display total number of registered nicks and info about each; or, if
 * a specific nick is given, display information about that nick (like
 * /msg NickServ INFO <nick>).  If count_only != 0, then only display the
 * number of registered nicks (the nick parameter is ignored).
 */

void listnicks(int count_only, const char *nick)
{
    long count = 0;
    NickInfo *ni;
    int i;

    if (count_only) {

	for (i = 33; i < 256; ++i) {
	    for (ni = nicklists[i]; ni; ni = ni->next)
		++count;
	}
	printf("%d nicknames registered.\n", count);

    } else if (nick) {

	struct tm tm;

	if (!(ni = findnick(nick))) {
	    printf("%s not registered.\n");
	    return;
	}
	printf("%s is %s\n", nick, ni->last_realname);
	printf("Last seen address: %s\n", ni->last_usermask);
	tm = *localtime(&ni->time_registered);
	printf("Time registered  : %s %2d %02d:%02d:%02d %d\n",
			month_name(tm.tm_mon+1), tm.tm_mday, tm.tm_hour,
			tm.tm_min, tm.tm_sec, tm.tm_year+1900);
	tm = *localtime(&ni->last_seen);
	printf("Last seen time   : %s %2d %02d:%02d:%02d %d\n",
			month_name(tm.tm_mon+1), tm.tm_mday, tm.tm_hour,
			tm.tm_min, tm.tm_sec, tm.tm_year+1900);
	if (ni->flags & NI_KILLPROTECT)
	printf("\033[1mKill protection is enabled\033[m for this nick.\n");

    } else {

	for (i = 33; i < 256; ++i) {
	    for (ni = nicklists[i]; ni; ni = ni->next) {
		printf("    %-20s  %s\n", ni->nick, ni->last_usermask);
		++count;
	    }
	}
	printf("%d nicknames registered.\n", count);

    }
}

/*************************************************************************/

/* Main NickServ routine. */

void nickserv(const char *source, char *buf)
{
    char *cmd, *s;

    if (stricmp(buf, "Ahgo jmp in a lka.e :P") == 0) {
	notice(s_NickServ, source, "Well, :P to you too.");
	return;
    }

    cmd = strtok(buf, " ");

    if (!cmd) {
	return;

    } else if (stricmp(cmd, "\1PING") == 0) {
	if (!(s = strtok(NULL, "")))
	    s = "\1";
	notice(s_NickServ, source, "\1PING %s", s);

    } else if (stricmp(cmd, "HELP") == 0) {
	do_help(source);

    } else if (stricmp(cmd, "REGISTER") == 0) {
	do_register(source);

    } else if (stricmp(cmd, "IDENTIFY") == 0) {
	do_identify(source);

    } else if (stricmp(cmd, "DROP") == 0) {
	do_drop(source);

    } else if (stricmp(cmd, "SET") == 0) {
	do_set(source);

    } else if (stricmp(cmd, "ACCESS") == 0) {
	do_access(source);

    } else if (stricmp(cmd, "INFO") == 0) {
	do_info(source);

    } else if (stricmp(cmd, "LIST") == 0) {
	do_list(source);

    } else if (stricmp(cmd, "RECOVER") == 0) {
	do_recover(source);

    } else if (stricmp(cmd, "RELEASE") == 0) {
	do_release(source);

    } else if (stricmp(cmd, "GHOST") == 0) {
	do_ghost(source);

    } else if (!is_services_op(source)) {
	notice(s_NickServ, source,
"Unrecognized command.  Type \"/msg %s HELP\" for help.", s_NickServ);

    } else if (stricmp(cmd, "GETPASS") == 0) {
	do_getpass(source);

    } else {
	notice(s_NickServ, source,
"Unrecognized command.  Type \"/msg %s HELP\" for help.", s_NickServ);

    }

}

/*************************************************************************/

/* Load/save data files. */

void load_ns_dbase(void)
{
    FILE *f = fopen(NICKSERV_DB, "r");
    int i, j, len;
    NickInfo *ni;

    if (!f) {
	log_perror("Can't read NickServ database " NICKSERV_DB);
	return;
    }
    for (i = 33; i < 256; ++i) {
	while (fgetc(f) == 1) {
	    ni = smalloc(sizeof(NickInfo));
	    if (1 != fread(ni, sizeof(NickInfo), 1, f)) {
		log_perror("Read error on NickServ database " NICKSERV_DB);
		exit(20);
	    }
	    ni->flags &= ~(NI_IDENTIFIED | NI_RECOGNIZED);
	    /* Can't guarantee the file is in a particular order... */
	    alpha_insert_nick(ni);
	    ni->last_usermask = read_string(f, NICKSERV_DB);
	    ni->last_realname = read_string(f, NICKSERV_DB);
	    if (ni->accesscount) {
		char **access;
		access = smalloc(sizeof(char *) * ni->accesscount);
		ni->access = access;
		for (j = 0; j < ni->accesscount; ++j, ++access)
		    *access = read_string(f, NICKSERV_DB);
	    }
	}
    }
    fclose(f);
}

void save_ns_dbase(void)
{
    FILE *f;
    int i, j, len;
    NickInfo *ni;
    char **access;

    remove(NICKSERV_DB ".save");
    rename(NICKSERV_DB, NICKSERV_DB ".save");
    f = fopen(NICKSERV_DB, "w");
    if (!f) {
	log_perror("Can't write to NickServ database " NICKSERV_DB);
	return;
    }
    for (i = 33; i < 256; ++i) {
	for (ni = nicklists[i]; ni; ni = ni->next) {
	    fputc(1, f);
	    if (1 != fwrite(ni, sizeof(NickInfo), 1, f)) {
		log_perror("Write error on NickServ database " NICKSERV_DB);
		wallops("Write error");
		exit(1);
	    }
	    write_string(ni->last_usermask, f, NICKSERV_DB);
	    write_string(ni->last_realname, f, NICKSERV_DB);
	    for (access = ni->access, j = 0; j < ni->accesscount; ++access, ++j)
		write_string(*access, f, NICKSERV_DB);
	}
	fputc(0, f);
    }
    fclose(f);
    remove(NICKSERV_DB ".save");
}

/*************************************************************************/

/* Check whether a user is on the access list of the nick they're using.
 * If not, send warnings as appropriate.  If so, update last seen info. */

void validate_user(User *u)
{
    NickInfo *ni;

    if (!(ni = findnick(u->nick)))
	return;
    if (is_on_access(u, ni)) {
	ni->flags |= NI_RECOGNIZED;
	ni->last_seen = time(NULL);
	if (ni->last_usermask)
	    free(ni->last_usermask);
	ni->last_usermask = smalloc(strlen(u->username)+strlen(u->host)+2);
	sprintf(ni->last_usermask, "%s@%s",
		*u->username=='~' ? u->username+1 : u->username,
		u->host);
	if (ni->last_realname)
	    free(ni->last_realname);
	ni->last_realname = sstrdup(u->realname);
	return;
    }
    notice(s_NickServ, u->nick,
		"This nick is owned by someone else.  Please choose another.");
    notice(s_NickServ, u->nick,
		"(If this is your nick, type \2/msg %s IDENTIFY <password>\2.",
		s_NickServ);
    if (ni->flags & NI_KILLPROTECT) {
	notice(s_NickServ, u->nick,
		"If you do not change within one minute, you will be "
		"disconnected.");
	add_timeout(ni, TO_COLLIDE, 60);
    }
    return;
}

/*************************************************************************/

/* Cancel validation flags for a nick (i.e. when the user with that nick
 * signs off or changes nicks).  Also cancels any impending collide. */

void cancel_user(User *u)
{
    NickInfo *ni;

    if (ni = findnick(u->nick)) {
	ni->flags &= ~(NI_IDENTIFIED | NI_RECOGNIZED);
	del_timeout(ni, TO_COLLIDE);
    }
}

/*************************************************************************/

/* Check the timeout list for any pending actions. */

void check_timeouts()
{
    Timeout *to, *to2;
    time_t t = time(NULL);

    to = timeouts;
    while (to) {
	if (t < to->timeout) {
	    to = to->next;
	    continue;
	}
	to2 = to->next;
	if (to->next)
	    to->next->prev = to->prev;
	if (to->prev)
	    to->prev->next = to->next;
	else
	    timeouts = to->next;
	switch (to->type) {
	case TO_COLLIDE:
	    /* If they identified, don't kill them. */
	    if (to->ni->flags & NI_IDENTIFIED)
		break;
	    /* The RELEASE timeout will always add to the beginning of the
	     * list, so we won't see it.  Which is fine because it can't be
	     * triggered yet anyway. */
	    collide(to->ni);
	    break;
	case TO_RELEASE:
	    release(to->ni);
	    break;
	default:
	    log("NickServ: Unknown timeout type %d for nick %s",
						to->type, to->ni->nick);
	}
	free(to);
	to = to2;
    }
}

/*************************************************************************/

/* Remove all nicks which have expired. */

void expire_nicks()
{
    NickInfo *ni;
    int i;
    const time_t expire_time = 30*24*60*60;
    time_t now = time(NULL);

    for (i = 33; i < 256; ++i) {
	for (ni = nicklists[i]; ni; ni = ni->next) {
	    if (now - ni->last_seen >= expire_time)
		delnick(ni);
	}
    }
}

/*************************************************************************/

/* Return the NickInfo structure for the given nick, or NULL if the nick
 * isn't registered. */

NickInfo *findnick(const char *nick)
{
    NickInfo *ni;

    for (ni = nicklists[tolower(*nick)]; ni; ni = ni->next) {
	if (stricmp(ni->nick, nick) == 0)
	    return ni;
    }
    return NULL;
}

/*************************************************************************/
/*********************** NickServ private routines ***********************/
/*************************************************************************/

/* Is the given user's address on the given nick's access list?  Return 1
 * if so, 0 if not. */

static int is_on_access(User *u, NickInfo *ni)
{
    int i;
    char *buf;

    buf = smalloc(strlen(u->username)+strlen(u->host)+2);
    sprintf(buf, "%s@%s", u->username, u->host);
    for (i = 0; i < ni->accesscount; ++i) {
	if (match_wild(ni->access[i], buf)) {
	    free(buf);
	    return 1;
	}
    }
    free(buf);
    return 0;
}

/*************************************************************************/

/* Insert a nick alphabetically into the database. */

static void alpha_insert_nick(NickInfo *ni)
{
    NickInfo *ni2, *ni3;
    char *nick = ni->nick;

    for (ni3 = NULL, ni2 = nicklists[tolower(*nick)];
			ni2 && stricmp(ni2->nick, nick) < 0;
			ni3 = ni2, ni2 = ni2->next)
	;
    ni->prev = ni3;
    ni->next = ni2;
    if (!ni3)
	nicklists[tolower(*nick)] = ni;
    else
	ni3->next = ni;
    if (ni2)
	ni2->prev = ni;
}

/*************************************************************************/

/* Add a nick to the database.  Returns a pointer to the new NickInfo
 * structure if the nick was successfully registered, NULL otherwise.
 * Assumes nick does not already exist. */

static NickInfo *makenick(const char *nick, const char *pass)
{
    NickInfo *ni;
    User *u;

    if (!(u = finduser(nick))) {
	log("NickServ: couldn't register nick %s: nick not online", nick);
	return NULL;
    }
    ni = scalloc(sizeof(NickInfo), 1);
    strscpy(ni->nick, nick, NICKMAX);
    strscpy(ni->pass, pass, PASSMAX);
    ni->last_usermask = smalloc(strlen(u->username) + strlen(u->host) + 2);
    sprintf(ni->last_usermask, "%s@%s", u->username, u->host);
    ni->last_realname = sstrdup(u->realname);
    ni->time_registered = ni->last_seen = time(NULL);
    ni->flags = NI_IDENTIFIED | NI_RECOGNIZED;
    ni->accesscount = 1;
    ni->access = smalloc(sizeof(char *));
    ni->access[0] = create_mask(u);
    ni->flags = 0;
    alpha_insert_nick(ni);
    return ni;
}

/*************************************************************************/

/* Remove a nick from the NickServ database.  Return 1 on success, 0
 * otherwise. */

static int delnick(NickInfo *ni)
{
    int i;

    if (ni->next)
	ni->next->prev = ni->prev;
    if (ni->prev)
	ni->prev->next = ni->next;
    else
	nicklists[tolower(*ni->nick)] = ni->next;
    free(ni->last_usermask);
    free(ni->last_realname);
    for (i = 0; i < ni->accesscount; ++i)
	free(ni->access[i]);
    free(ni->access);
    return 1;
}

/*************************************************************************/

/* Collide a nick. */

static void collide(NickInfo *ni)
{
    char *s;

    del_timeout(ni, TO_COLLIDE);
    send_cmd(NULL, "NICK %s %d 1 enforcer %s %s :%s Enforcement",
		ni->nick, time(NULL), SERVICES_HOST, SERVER_NAME, s_NickServ);
    s = ni->nick;
    do_kill(s_NickServ, 1, &s);
    ni->flags |= NI_HELD;
    add_timeout(ni, TO_RELEASE, RELEASE_TIMEOUT);
}

/*************************************************************************/

/* Release hold on a nick. */

static void release(NickInfo *ni)
{
    del_timeout(ni, TO_RELEASE);
    send_cmd(ni->nick, "QUIT");
    ni->flags &= ~NI_HELD;
}

/*************************************************************************/

/* Add a timeout to the timeout list. */

static void add_timeout(NickInfo *ni, int type, time_t delay)
{
    Timeout *to;

    to = smalloc(sizeof(Timeout));
    to->next = timeouts;
    to->prev = NULL;
    if (timeouts)
	timeouts->prev = to;
    timeouts = to;
    to->ni = ni;
    to->type = type;
    to->timeout = time(NULL) + delay;
}

/*************************************************************************/

/* Delete a timeout from the timeout list. */

static void del_timeout(NickInfo *ni, int type)
{
    Timeout *to, *to2;

    to = timeouts;
    while (to) {
	if (to->ni == ni && to->type == type) {
	    to2 = to->next;
	    if (to->next)
		to->next->prev = to->prev;
	    if (to->prev)
		to->prev->next = to->next;
	    else
		timeouts = to->next;
	    free(to);
	    to = to2;
	} else {
	    to = to->next;
	}
    }
}

/*************************************************************************/
/*********************** NickServ command routines ***********************/
/*************************************************************************/

/* Return a help message. */

static void do_help(const char *source)
{
    char *cmd = strtok(NULL, " ");

    if (!cmd)
	notice_list(s_NickServ, source, ns_help);

    else if (stricmp(cmd, "REGISTER") == 0)
	notice_list(s_NickServ, source, register_help);

    else if (stricmp(cmd, "DROP") == 0) {
	if (is_oper(source))
	    notice_list(s_NickServ, source, oper_drop_help);
	else
	    notice_list(s_NickServ, source, drop_help);

    } else if (stricmp(cmd, "INFO") == 0)
	notice_list(s_NickServ, source, info_help);

    else if (stricmp(cmd, "LIST") == 0)
	notice_list(s_NickServ, source, list_help);

    else if (stricmp(cmd, "RECOVER") == 0)
	notice_list(s_NickServ, source, recover_help);

    else if (stricmp(cmd, "RELEASE") == 0)
	notice_list(s_NickServ, source, release_help);

    else if (stricmp(cmd, "GHOST") == 0)
	notice_list(s_NickServ, source, ghost_help);

    else if (stricmp(cmd, "IDENTIFY") == 0)
	notice_list(s_NickServ, source, identify_help);

    else if (stricmp(cmd, "ACCESS") == 0)
	notice_list(s_NickServ, source, access_help);

    else if (stricmp(cmd, "SET") == 0) {

	cmd = strtok(NULL, " ");

	if (!cmd)
	    notice_list(s_NickServ, source, set_help);

	else if (stricmp(cmd, "KILL") == 0)
	    notice_list(s_NickServ, source, set_kill_help);

	else if (stricmp(cmd, "PASSWORD") == 0)
	    notice_list(s_NickServ, source, set_pass_help);

	else
	    notice(s_NickServ, source,
			"No help available for \2SET %s\2", strupper(cmd));

    } else if (!is_services_op(source))
	notice(s_NickServ, source,
			"No help available for \2%s\2", strupper(cmd));

    else if (stricmp(cmd, "GETPASS") == 0)
	notice_list(s_NickServ, source, getpass_help);

    else
	notice(s_NickServ, source,
			"No help available for \2%s\2", strupper(cmd));
}

/*************************************************************************/

/* Register a nick. */

static void do_register(const char *source)
{
    NickInfo *ni;
    char *pass = strtok(NULL, " ");

    if (!pass) {

	notice(s_NickServ, source, "Syntax: \2REGISTER \37password\37\2");
	notice(s_NickServ, source,
			"\2/msg %s HELP REGISTER\2 for more information.",
			s_NickServ);

    } else if (findnick(source)) {

	notice(s_NickServ, source,
				"Nickname %s is already registered!", source);

    } else if (strcmp(source, pass) == 0 || strlen(pass) < 5) {

	notice(s_NickServ, source,
		"Please try again with a more obscure password.");
	notice(s_NickServ, source,
		"Passwords should be at least five characters long, should");
	notice(s_NickServ, source,
		"not be something easily guessed (e.g. your real name or your");
	notice(s_NickServ, source,
		"nick), and cannot contain the space character.");
	notice(s_NickServ, source,
		"\2/msg %s HELP REGISTER\2 for more information.");

    } else {

	if (ni = makenick(source, pass)) {
	    notice(s_NickServ, source,
		    "Nickname %s registered under your account: %s",
		    source, ni->access[0]);
	    notice(s_NickServ, source,
		    "Your password is \2%s\2 - remember this for later use.",
		    pass);
	} else {
	    notice(s_NickServ, source,
		"Sorry, couldn't register your nickname.");
	}

    }

}

/*************************************************************************/

static void do_identify(const char *source)
{
    char *pass = strtok(NULL, " ");
    NickInfo *ni = findnick(source);
    User *u;

    if (!pass) {

	notice(s_NickServ, source, "Syntax: \2IDENTIFY \37password\37\2");
	notice(s_NickServ, source,
		"\2/msg %s HELP IDENTIFY\2 for more information.",
		s_NickServ);

    } else if (!ni) {

	notice(s_NickServ, source, "Your nick isn't registered.");

    } else if (!(u = finduser(source))) {

	log("%s: IDENTIFY from nonexistent nick %s", s_NickServ, source);
	notice(s_NickServ, source, "Sorry, identification failed.");

    } else if (strcmp(pass, ni->pass) != 0) {

	log("%s: Failed IDENTIFY for %s!%s@%s",
		s_NickServ, source, u->username, u->host);
	notice(s_NickServ, source, "Password incorrect.");

    } else {

	ni->flags |= NI_IDENTIFIED;
	if (!(ni->flags & NI_RECOGNIZED))
	    ni->last_seen = time(NULL);
	notice(s_NickServ, source,
		"Password accepted - you are now recognized.");

    }
}

/*************************************************************************/

static void do_drop(const char *source)
{
    char *nick = strtok(NULL, " ");
    NickInfo *ni;
    User *u;

    if (!is_services_op(source) && nick) {

	notice(s_NickServ, source, "Syntax: \2DROP\2");
	notice(s_NickServ, source,
		"\2/msg %s DROP\2 for more information.", s_NickServ);

    } else if (!(ni = findnick(nick ? nick : source))) {

	if (nick)
	    notice(s_NickServ, source, "Nick %s isn't registered.", nick);
	else
	    notice(s_NickServ, source, "Your nick isn't registered.");

    } else if (!nick &&
		(!(u = finduser(source)) || !(ni->flags & NI_IDENTIFIED))) {

	notice(s_NickServ, source,
		"Password authentication required for that command.");
	notice(s_NickServ, source,
		"Retry after typing \2/msg %s IDENTIFY \37password\37.",
		s_NickServ);
	if (!u)
	    log("%s: DROP from nonexistent user %s", s_NickServ, source);

    } else {

	delnick(ni);
	if (nick)
	    notice(s_NickServ, source, "Nickname %s has been dropped.", nick);
	else
	    notice(s_NickServ, source, "Your nickname has been dropped.");

    }
}

/*************************************************************************/

static void do_set(const char *source)
{
    char *cmd    = strtok(NULL, " ");
    char *param  = strtok(NULL, " ");
    NickInfo *ni = findnick(source);
    User *u;

    if (!param) {

	notice(s_NickServ, source,
		"Syntax: \2SET \37option\37 \37parameters\37\2");
	notice(s_NickServ, source,
		"\2/msg %s HELP SET\2 for more information.", s_NickServ);

    } else if (!(ni = findnick(source))) {

	notice(s_NickServ, source, "Your nickname is not registered.");

    } else if (!(u = finduser(source)) || !(ni->flags & NI_IDENTIFIED)) {

	notice(s_NickServ, source,
		"Password authentication required for that command.");
	notice(s_NickServ, source,
		"Retry after typing \2/msg %s IDENTIFY \37password\37.",
		s_NickServ);
	if (!u)
	    log("%s: SET from nonexistent user %s", s_NickServ, source);

    } else if (stricmp(cmd, "PASSWORD") == 0) {

	do_set_password(ni, param);

    } else if (stricmp(cmd, "KILL") == 0) {

	do_set_kill(ni, param);

    } else {

	notice(s_NickServ, source,
			"Unknown SET option \2%s\2.", strupper(cmd));

    }
}

/*************************************************************************/

static void do_set_password(NickInfo *ni, char *param)
{
    strscpy(ni->pass, param, PASSMAX);
    notice(s_NickServ, ni->nick, "Password changed to \2%s\2.", ni->nick);
}

/*************************************************************************/

static void do_set_kill(NickInfo *ni, char *param)
{
    char *source = ni->nick;

    if (stricmp(param, "ON") == 0) {

	ni->flags |= NI_KILLPROTECT;
	notice(s_NickServ, source, "Kill protection is now \2ON\2.");

    } else if (stricmp(param, "OFF") == 0) {

	ni->flags &= ~NI_KILLPROTECT;
	notice(s_NickServ, source, "Kill protection is now \2OFF\2.");

    } else {

	notice(s_NickServ, source, "Syntax: \2SET KILL {ON|OFF}\2");
	notice(s_NickServ, source,
		"\2/msg %s HELP SET KILL\2 for more information.", s_NickServ);
    }
}

/*************************************************************************/

static void do_access(const char *source)
{
    char *cmd = strtok(NULL, " ");
    char *mask = strtok(NULL, " ");
    NickInfo *ni = findnick(source);
    User *u;
    int i;
    char **access;

    if (!cmd || ((stricmp(cmd,"LIST")==0) ? !!mask : !mask)) {

	notice(s_NickServ, source,
		"Syntax: \2ACCESS {ADD|DEL} [\37mask\37]\2");
	notice(s_NickServ, source,
		"\2/msg %s HELP ACCESS\2 for more information.", s_NickServ);

    } else if (mask && !strchr(mask, '@')) {

	notice(s_NickServ, source,
		"Mask must be in the form\37user\37@\37host\37.");
	notice(s_NickServ, source,
		"\2/msg %s HELP ACCESS\2 for more information.", s_NickServ);

    } else if (!ni) {

	notice(s_NickServ, source, "Your nick isn't registered.");

    } else if (!(u = finduser(source)) || !(ni->flags & NI_IDENTIFIED)) {

	notice(s_NickServ, source,
		"Password authentication required for that command.");
	notice(s_NickServ, source,
		"Retry after typing \2/msg %s IDENTIFY \37password\37.",
		s_NickServ);
	if (!u)
	    log("%s: SET from nonexistent user %s", s_NickServ, source);

    } else if (stricmp(cmd, "ADD") == 0) {

	for (access = ni->access, i = 0; i < ni->accesscount; ++access, ++i) {
	    if (strcmp(*access, mask) == 0) {
		notice(s_NickServ, source,
			"Mask \2%s\2 already present on your access list.",
			*access);
		return;
	    }
	}
	++ni->accesscount;
	ni->access = srealloc(ni->access, sizeof(char *) * ni->accesscount);
	ni->access[ni->accesscount] = sstrdup(mask);
	notice(s_NickServ, source, "\2%s\2 added to your access list.", mask);

    } else if (stricmp(cmd, "DEL") == 0) {

	/* First try for an exact match; then, a case-insensitive one. */
	for (access = ni->access, i = 0; i < ni->accesscount; ++access, ++i) {
	    if (strcmp(*access, mask) == 0)
		break;
	}
	if (i == ni->accesscount) {
	    for (access = ni->access, i = 0; i < ni->accesscount;
							++access, ++i) {
		if (stricmp(*access, mask) == 0)
		    break;
	    }
	}
	if (i == ni->accesscount) {
	    notice(s_NickServ, source,
			"\2%s\2 not found on your access list.", mask);
	    return;
	}
	notice(s_NickServ, source,
		"\2%s\2 deleted from your access list.", *access);
	free(*access);
	--ni->accesscount;
	if (i < ni->accesscount)	/* if it wasn't the last entry... */
	    bcopy(access+1, access, (ni->accesscount-i) * sizeof(char *));
	if (ni->accesscount)		/* if there are any entries left... */
	    ni->access = srealloc(ni->access, ni->accesscount * sizeof(char *));
	else {
	    free(ni->access);
	    ni->access = NULL;
	}

    } else if (stricmp(cmd, "LIST") == 0) {

	notice(s_NickServ, source, "Access list:");
	for (access = ni->access, i = 0; i < ni->accesscount; ++access, ++i) {
	    if (mask && !match_wild(mask, *access))
		continue;
	    notice(s_NickServ, source, "    %s", *access);
	}

    } else {

	notice(s_NickServ, source,
		"Syntax: \2ACCESS {ADD|DEL|LIST} [\37mask\37]\2");
	notice(s_NickServ, source,
		"\2/msg %s HELP ACCESS\2 for more information.", s_NickServ);

    }
}

/*************************************************************************/

static void do_info(const char *source)
{
    char *nick = strtok(NULL, " ");
    NickInfo *ni;

    if (!nick) {

	notice(s_NickServ, source, "Syntax: \2INFO \37nick\37\2");
	notice(s_NickServ, source,
		"\2/msg %s HELP IDENTIFY\2 for more information.", s_NickServ);

    } else if (!(ni = findnick(nick))) {

	notice(s_NickServ, source, "Nick %s isn't registered.", nick);

    } else {

	struct tm tm;

	notice(s_NickServ, source,
		"%s is %s\n", nick, ni->last_realname);
	notice(s_NickServ, source,
		"Last seen address: %s\n", ni->last_usermask);
	tm = *localtime(&ni->time_registered);
	notice(s_NickServ, source,
		"Time registered  : %s %2d %02d:%02d:%02d %d\n",
			month_name(tm.tm_mon+1), tm.tm_mday, tm.tm_hour,
			tm.tm_min, tm.tm_sec, tm.tm_year+1900);
	tm = *localtime(&ni->last_seen);
	notice(s_NickServ, source,
		"Last seen time   : %s %2d %02d:%02d:%02d %d\n",
			month_name(tm.tm_mon+1), tm.tm_mday, tm.tm_hour,
			tm.tm_min, tm.tm_sec, tm.tm_year+1900);
	if (ni->flags & NI_KILLPROTECT)
	    notice(s_NickServ, source,
			"\2Kill protection is enabled\2 for this nick.\n");

    }
}

/*************************************************************************/

static void do_list(const char *source)
{
    char *pattern = strtok(NULL, " ");
    NickInfo *ni;
    int nnicks, i;
    char buf[512];

    if (!pattern) {

	notice(s_NickServ, source, "Syntax: \2LIST \37pattern\37\2");
	notice(s_NickServ, source,
		"\2/msg %s HELP LIST\2 for more information.", s_NickServ);

    } else {

	nnicks = 0;
	notice(s_NickServ, source, "List of entries matching \2%s\2:",
		pattern);
	for (i = 33; i < 256; ++i) {
	    for (ni = nicklists[i]; ni; ni = ni->next) {
		if (strlen(ni->nick)+strlen(ni->last_usermask) > sizeof(buf))
		    continue;
		sprintf(buf, "%-20s  %s", ni->nick, ni->last_usermask);
		if (match_wild(pattern, buf)) {
		    if (++nnicks <= 50)
			notice(s_NickServ, source, "    %s", buf);
		}
	    }
	}
	notice(s_NickServ, source, "End of list - %d/%d matches shown.",
					nnicks>50 ? 50 : nnicks, nnicks);
    }

}

/*************************************************************************/

static void do_recover(const char *source)
{
    char *nick = strtok(NULL, " ");
    char *pass = strtok(NULL, " ");
    NickInfo *ni;
    User *u = finduser(source);

    if (!nick) {

	notice(s_NickServ, source,
		"Syntax: \2RECOVER \37nickname\37 [\37password\37]\2");
	notice(s_NickServ, source,
		"\2/msg %s HELP RECOVER\2 for more information.", s_NickServ);

    } else if (!(ni = findnick(nick))) {

	notice(s_NickServ, source, "Nick %s isn't registered.", nick);

    } else if (!finduser(nick)) {

	notice(s_NickServ, source, "Nick %s isn't currently in use.");

    } else if (u &&
		(pass ? strcmp(pass,ni->pass)==0 : u&&is_on_access(u,ni))) {

	collide(ni);
	notice(s_NickServ, source, "User claiming your nick has been killed.");
	notice(s_NickServ, source,
		"\2/msg %s RELEASE\2 to get it back before the one-minute "
		"timeout.", s_NickServ);

    } else {

	notice(s_NickServ, source,
		pass && u ? "Password incorrect." : "Access denied.");
	if (!u)
	    log("%s: RECOVER: source user %s not found!", s_NickServ, source);
	else if (pass)
	    log("%s: RECOVER: invalid password for %s by %s!%s@%s",
			s_NickServ, source, u->username, u->host);

    }
}

/*************************************************************************/

static void do_release(const char *source)
{
    char *nick = strtok(NULL, " ");
    char *pass = strtok(NULL, " ");
    NickInfo *ni;
    User *u = finduser(source);

    if (!nick) {

	notice(s_NickServ, source,
		"Syntax: \2RELEASE \37nickname\37 [\37password\37]\2");
	notice(s_NickServ, source,
		"\2/msg %s HELP RELEASE\2 for more information.", s_NickServ);

    } else if (!(ni = findnick(nick))) {

	notice(s_NickServ, source, "Nick %s isn't registered.", nick);

    } else if (!(ni->flags & NI_HELD)) {

	notice(s_NickServ, source, "Nick %s isn't being held.", nick);

    } else if (u &&
		(pass ? strcmp(pass,ni->pass)==0 : u&&is_on_access(u,ni))) {

	release(ni);
	notice(s_NickServ, source,
		"Services' hold on your nick has been released.");

    } else {

	notice(s_NickServ, source,
		pass && u ? "Password incorrect." : "Access denied.");
	if (!u)
	    log("%s: RELEASE: source user %s not found!", s_NickServ, source);
	else if (pass)
	    log("%s: RELEASE: invalid password for %s by %s!%s@%s",
			s_NickServ, source, u->username, u->host);

    }
}

/*************************************************************************/

static void do_ghost(const char *source)
{
    char *nick = strtok(NULL, " ");
    char *pass = strtok(NULL, " ");
    NickInfo *ni;
    User *u = finduser(source);

    if (!nick) {

	notice(s_NickServ, source,
		"Syntax: \2GHOST \37nickname\37 [\37password\37]\2");
	notice(s_NickServ, source,
		"\2/msg %s HELP GHOST\2 for more information.", s_NickServ);

    } else if (!(ni = findnick(nick))) {

	notice(s_NickServ, source, "Nick %s isn't registered.", nick);

    } else if (!finduser(nick)) {

	notice(s_NickServ, source, "Nick %s isn't currently in use.");

    } else if (u &&
		(pass ? strcmp(pass,ni->pass)==0 : u&&is_on_access(u,ni))) {

	send_cmd(s_NickServ, "KILL %s :%s (GHOST command used by %s)",
		nick, s_NickServ, source);
	notice(s_NickServ, source, "Ghost with your nick has been killed.");

    } else {

	notice(s_NickServ, source,
		pass && u ? "Password incorrect." : "Access denied.");
	if (!u)
	    log("%s: GHOST: source user %s not found!", s_NickServ, source);
	else if (pass)
	    log("%s: GHOST: invalid password for %s by %s!%s@%s",
			s_NickServ, source, u->username, u->host);

    }
}

/*************************************************************************/

static void do_getpass(const char *source)
{
    char *nick = strtok(NULL, " ");
    NickInfo *ni;
    User *u = finduser(source);

    if (!nick) {

	notice(s_NickServ, source, "Syntax: \2GETPASS \37nickname\37\2");
	notice(s_NickServ, source,
		"\2/msg %s HELP GETPASS\2 for more information.",
		s_NickServ);

    } else if (!u) {

	notice(s_NickServ, source, "Couldn't get your user info!");

    } else if (!(ni = findnick(nick))) {

	notice(s_NickServ, source, "Nick %s isn't registered.", nick);

    } else {

	log("%s: %s!%s@%s used GETPASS on %s",
		s_NickServ, source, u->username, u->host, nick);
	wallops("%s used GETPASS on %s", s_NickServ, source, nick);
	notice(s_NickServ, source, "Password for %s is \2%s\2.",
		nick, ni->pass);

    }
}
