/* Routines to maintain list of online users */

#include "services.h"

static User *userlist = NULL;

/*************************************************************************/

/* Send the current list of users to the named user. */
void send_user_list(const char *user)
{
    User *u = userlist;

    notice(s_OperServ, user, "User list:");
    while (u) {
	notice(s_OperServ, user, "%s!%s@%s +%s%s%s", u->nick, u->username,
		u->host, (u->mode&UMODE_O)?"o":"", (u->mode&UMODE_I)?"i":"",
		(u->mode&UMODE_W)?"w":"");
	u = u->next;
    }
}

/*************************************************************************/

/* Find a user by nick.  Return NULL if user could not be found. */
User *finduser(const char *nick)
{
    User *user = userlist;
    while (user && stricmp(user->nick, nick) != 0)
	user = user->next;
    return user;
}

/*************************************************************************/

/* Handle a server NICK command. */
void do_nick(const char *source, int ac, char **av)
{
    User *user;

    if (!*source) {

	/* First check for AKILLs. */
	check_akill(av[0], av[3], av[4]);

	/* Allocate pre-cleared block of memory */
	user = scalloc(sizeof(User), 1);
	user->next = userlist;
	if (userlist)
	    userlist->prev = user;
	userlist = user;
	strscpy(user->nick, av[0], sizeof(user->nick));
	user->signon = atol(av[2]);
	user->username = sstrdup(av[3]);
	user->host = sstrdup(av[4]);
	strlower(user->host);
	user->server = sstrdup(av[5]);
	user->realname = sstrdup(av[6]);

    } else {

	user = finduser(source);
	if (!user) {
	    log("user: NICK from nonexistent nick %s: %s", source,
							merge_args(ac, av));
	    return;
	}
	cancel_user(user);
	strscpy(user->nick, av[0], sizeof(user->nick));

    }

    validate_user(user);
}

/*************************************************************************/

/* Handle a JOIN command. */
void do_join(const char *source, int ac, char **av)
{
    User *user;
    char *s, *t;
    struct u_chanlist *c;

    user = finduser(source);
    if (!user) {
	log("user: JOIN from nonexistent user %s: %s", source,
							merge_args(ac, av));
	return;
    }
    t = av[0];
    while (*(s=t)) {
	t = s + strcspn(s, ",");
	if (*t)
	    *t++ = 0;
	if (check_kick(user, s))
	    continue;
/*notice(s_OperServ,"Alcan",":%s JOIN %s",user->nick,s);*/
	chan_adduser(user, s);
/*send_channel_list("Alcan");*/
	c = smalloc(sizeof(*c));
	c->next = user->chans;
	c->prev = NULL;
	if (user->chans)
	    user->chans->prev = c;
	user->chans = c;
	c->chan = findchan(s);
    }
}

/*************************************************************************/

/* Handle a PART command. */
void do_part(const char *source, int ac, char **av)
{
    User *user;
    char *s, *t;
    struct u_chanlist *c;

    user = finduser(source);
    if (!user) {
	log("user: PART from nonexistent user %s: %s", source,
							merge_args(ac, av));
	return;
    }
    t = av[0];
    while (*(s=t)) {
	t = s + strcspn(s, ",");
	if (*t)
	    *t++ = 0;
	for (c = user->chans; c && stricmp(s, c->chan->name) != 0; c = c->next)
	    ;
	if (c) {
/*notice(s_OperServ,"Alcan",":%s PART %s",user->nick,s);*/
	    chan_deluser(user, c->chan);
/*send_channel_list("Alcan");*/
	    if (c->next)
		c->next->prev = c->prev;
	    if (c->prev)
		c->prev->next = c->next;
	    else
		user->chans = c->next;
	    free(c);
	}
    }
}

/*************************************************************************/

/* Handle a KICK command. */
void do_kick(const char *source, int ac, char **av)
{
    User *user;
    char *s, *t;
    struct u_chanlist *c;

    t = av[1];
    while (*(s=t)) {
	t = s + strcspn(s, ",");
	if (*t)
	    *t++ = 0;
	user = finduser(s);
	if (!user) {
	    log("user: KICK for nonexistent user %s on %s: %s", s, av[0],
						merge_args(ac-2, av+2));
	    continue;
	}
	for (c = user->chans; c && stricmp(av[0], c->chan->name) != 0;
								c = c->next)
	    ;
	if (c) {
/*notice(s_OperServ,"Alcan","KICK %s %s",s,user->nick);*/
	    chan_deluser(user, c->chan);
/*send_channel_list("Alcan");*/
	    if (c->next)
		c->next->prev = c->prev;
	    if (c->prev)
		c->prev->next = c->next;
	    else
		user->chans = c->next;
	    free(c);
	}
    }
}

/*************************************************************************/

/* Handle a MODE command for a user. */
void do_umode(const char *source, int ac, char **av)
{
    User *user;
    char *s;
    int add = 1;		/* 1 if adding modes, 0 if deleting */

    if (stricmp(source, av[0]) != 0) {
	log("user: MODE %s %s from different nick %s!", av[0], av[1], source);
	wallops("%s attempted to change mode %s for %s", source, av[1], av[0]);
	return;
    }
    user = finduser(source);
    if (!user) {
	log("user: MODE %s for nonexistent nick %s: %s", av[1], source,
							merge_args(ac, av));
	return;
    }
    s = av[1];
    while (*s) {
	switch (*s++) {
	    case '+': add = 1; break;
	    case '-': add = 0; break;
	    case 'o': add ? (user->mode |= UMODE_O) : (user->mode &= ~UMODE_O); break;
	    case 'i': add ? (user->mode |= UMODE_I) : (user->mode &= ~UMODE_I); break;
	    case 'w': add ? (user->mode |= UMODE_W) : (user->mode &= ~UMODE_W); break;
	    case 's': add ? (user->mode |= UMODE_S) : (user->mode &= ~UMODE_S); break;
	}
    }
}

/*************************************************************************/

/* Handle a QUIT command. */
void do_quit(const char *source, int ac, char **av)
{
    User *user;
    struct u_chanlist *c, *c2;
    struct u_chaninfolist *ci, *ci2;

    user = finduser(source);
    if (!user) {
	log("user: QUIT from nonexistent user %s: %s", source,
							merge_args(ac, av));
	return;
    }
    cancel_user(user);
    free(user->username);
    free(user->host);
    free(user->realname);
    free(user->server);
    c = user->chans;
    while (c) {
	c2 = c->next;
/*notice(s_OperServ,"Alcan",":%s QUIT %s",user->nick,c->chan->name);*/
	chan_deluser(user, c->chan);
/*send_channel_list("Alcan");*/
	free(c);
	c = c2;
    }
    ci = user->founder_chans;
    while (ci) {
	ci2 = ci->next;
	free(ci);
	ci = ci2;
    }
    if (user->prev)
	user->prev->next = user->next;
    else
	userlist = user->next;
    if (user->next)
	user->next->prev = user->prev;
    free(user);
}

/*************************************************************************/

/* Handle a KILL command. */
void do_kill(const char *source, int ac, char **av)
{
    User *user;
    struct u_chanlist *c, *c2;
    struct u_chaninfolist *ci, *ci2;

    user = finduser(av[0]);
    if (!user)
	return;
    cancel_user(user);
    free(user->username);
    free(user->host);
    free(user->realname);
    free(user->server);
    c = user->chans;
    while (c) {
	c2 = c->next;
/*notice(s_OperServ,"Alcan",":%s KILL %s",user->nick,c->chan->name);*/
	chan_deluser(user, c->chan);
/*send_channel_list("Alcan");*/
	free(c);
	c = c2;
    }
    ci = user->founder_chans;
    while (ci) {
	ci2 = ci->next;
	free(ci);
	ci = ci2;
    }
    if (user->prev)
	user->prev->next = user->next;
    else
	userlist = user->next;
    if (user->next)
	user->next->prev = user->prev;
    free(user);
}

/*************************************************************************/

int is_oper(const char *nick)
{
    User *user = finduser(nick);
    return user && (user->mode & UMODE_O);
}

int is_on_chan(const char *nick, const char *chan)
{
    User *u = finduser(nick);
    struct u_chanlist *c;

    if (!u)
	return 0;
    for (c = u->chans; c; c = c->next) {
	if (stricmp(c->chan->name, chan) == 0)
	    return 1;
    }
    return 0;
}

int is_chanop(const char *nick, const char *chan)
{
    Channel *c = findchan(chan);
    struct c_userlist *u;

    if (!c)
	return 0;
    for (u = c->chanops; u; u = u->next) {
	if (stricmp(u->user->nick, nick) == 0)
	    return 1;
    }
    return 0;
}

int is_voiced(const char *nick, const char *chan)
{
    Channel *c = findchan(chan);
    struct c_userlist *u;

    if (!c)
	return 0;
    for (u = c->voices; u; u = u->next) {
	if (stricmp(u->user->nick, nick) == 0)
	    return 1;
    }
    return 0;
}

/*************************************************************************/

/* Does the user's usermask match the given mask (either nick!user@host or
 * just user@host)? */
int match_usermask(const char *mask, User *user)
{
    char *mask2 = sstrdup(mask);
    char *nick, *username, *host, *nick2, *host2;
    int result;

    if (strchr(mask2, '!')) {
	nick = strtok(mask2, "!");
	username = strtok(NULL, "@");
    } else {
	nick = NULL;
	username = strtok(mask2, "@");
    }
    host = strtok(NULL, "");
    if (!username || !host) {
	free(mask2);
	return 0;
    }
    nick2 = strlower(sstrdup(user->nick));
    host2 = strlower(sstrdup(user->host));
    result = (nick ? match_wild(strlower(nick), nick2) : 1) &&
		match_wild(username, user->username) &&
		match_wild(strlower(host), host2);
    free(mask2);
    free(nick2);
    free(host2);
    return result;
}

/*************************************************************************/

/* Split a usermask up into its constitutent parts.  Returned strings are
 * malloc()'d, and should be free()'d when done with.  Returns "*" for
 * missing parts. */
void split_usermask(const char *mask, char **nick, char **user, char **host)
{
    char *mask2 = sstrdup(mask);

    *nick = strtok(mask2, "!");
    *user = strtok(NULL, "@");
    *host = strtok(NULL, "");
    /* Handle special case: mask == user@host */
    if (*nick && !*user && strchr(*nick, '@')) {
	*nick = NULL;
	*user = strtok(mask2, "@");
	*host = strtok(NULL, "");
    }
    if (!*nick)
	*nick = "*";
    if (!*user)
	*user = "*";
    if (!*host)
	*host = "*";
    *nick = sstrdup(*nick);
    *user = sstrdup(*user);
    *host = sstrdup(*host);
    free(mask2);
}

/*************************************************************************/

/* Given a user, return a mask that will most likely match any address the
 * user will have from that location.  For IP addresses, wildcards the
 * appropriate subnet mask (e.g. 35.1.1.1 -> 35.*; 128.2.1.1 -> 128.2.*);
 * for named addresses, wildcards the leftmost part of the name unless the
 * name only contains two parts.  If the username begins with a ~, delete
 * it.  The returned character string is malloc'd and should be free'd
 * when done with.
 */
char *create_mask(User *u)
{
    char *mask, *s, *end;

    end = mask = smalloc(strlen(u->username) + strlen(u->host) + 2);
    end += sprintf(end, "%s@",
			(u->username[0] == '~') ? u->username+1 : u->username);
    if (strspn(u->host, "0123456789.") == strlen(u->host)) {	/* IP addr */
	if ((s = strchr(u->host, '.')) &&
		(s = strchr(s, '.')) &&
		(s = strchr(s, '.')) &&
		!strchr(s, '.')) {
	    strcpy(end, u->host);
	} else {
	    s = sstrdup(u->host);
	    *strrchr(s, '.') = 0;
	    if (atoi(u->host) < 192)
		*strrchr(s, '.') = 0;
	    if (atoi(u->host) < 128)
		*strrchr(s, '.') = 0;
	    sprintf(end, "%s.*", s);
	    free(s);
	}
    } else {
	if ((s = strchr(u->host, '.')) && strchr(s+1, '.')) {
	    s = sstrdup(strchr(u->host, '.')-1);
	    *s = '*';
	} else {
	    s = sstrdup(u->host);
	}
	strcpy(end, s);
	free(s);
    }
    return mask;
}
